#include <iostream>
#include <cstring> // Para usar strcpy()
#include <cctype> // Para usar isdigit() e isalpha()
#include <fstream> // Para usar ficheros
#include <sstream> // Para usar stringstream
#include <vector> // Para usar vector

using namespace std;

const int KMAXNIF=10;
const int KMAXNAME=50;
const int KMAXTELEPHONE=14;
const char KFILEPATIENTS[]="patients.bin";
const char KFILEANALYSIS[]="analysis.bin";
const char KFILEWRONGPATIENTS[]="wrong_patients.txt";
const char KFILESTATISTICS[]="statistics.txt";

// Registro para los pacientes
struct Patient{
    string nif;
    string name;
    string telephone;
    
};

// Registro para los pacientes en fichero binario
struct PatientBin{
    char nif[KMAXNIF];
    char name[KMAXNAME];
    char telephone[KMAXTELEPHONE];
};

// Registro para las fechas
struct Date{
    int day;
    int month;
    int year;
};

// Registro para las analíticas
struct Analysis{
    unsigned int id;
    char nif[KMAXNIF];
    Date dateAnalysis ;
    float weight;
    float height;
};

// Registro para la base de datos
struct Database{
    unsigned int nextId;
    vector<Patient> patients;
    vector<Analysis> analysis;
};

// Tipos de errores posibles
enum Error{
    ERR_ARGS,
    ERR_FILE,
    ERR_OPTION,
    ERR_PATIENT_EXISTS,
    ERR_PATIENT_NOT_EXISTS,
    ERR_WRONG_DATE,
    ERR_WRONG_NAME,
    ERR_WRONG_NIF,
    ERR_WRONG_NUMBER,
    ERR_WRONG_TELEPHONE
};

/*
Función que muestra los distintos tipos de errores posibles
e: tipo de error a mostrar
return: void
*/
void error(Error e){
    switch (e){
        case ERR_ARGS: cout << "ERROR: wrong arguments" << endl;
            break;
        case ERR_FILE: cout << "ERROR: cannot open file" << endl;
            break;
        case ERR_OPTION: cout << "ERROR: wrong option" << endl;
            break;
        case ERR_PATIENT_EXISTS: cout << "ERROR: patient already exists" << endl;
            break;
        case ERR_PATIENT_NOT_EXISTS: cout << "ERROR: patient does not exist" << endl;
            break;
        case ERR_WRONG_DATE: cout << "ERROR: wrong date" << endl;
            break;
        case ERR_WRONG_NAME: cout << "ERROR: wrong name " << endl;
            break;
        case ERR_WRONG_NIF: cout << "ERROR: wrong NIF " << endl;
            break;
        case ERR_WRONG_NUMBER: cout << "ERROR: wrong number " << endl;
            break;
        case ERR_WRONG_TELEPHONE: cout << "ERROR: wrong telephone number" << endl;
    }
}

/*
Función que comprueba si un NIF es correcto
nif: NIF a comprobar
return: true si es correcto, false en caso contrario
*/
bool checkNif(string nif){
    bool correct=true;
    
    if(nif.length()!=9){
        correct=false;
    }
    else{
        for(int i=0;i<8 && correct;i++){
            if(!isdigit(nif[i])){
                correct=false;
            }
        }
        if(correct && !isalpha(nif[8])){
            correct=false;
        }
    }
    
    return correct;
}

/*
Función que comprueba si un nombre es correcto
name: nombre a comprobar
return: true si es correcto, false en caso contrario
*/
bool checkName(string name){
    bool correct=true;
    
    if(name.length()<3){
        correct=false;
    }
    
    return correct;
}

/*
Función que comprueba si un teléfono es correcto
telephone: teléfono a comprobar
return: true si es correcto, false en caso contrario
*/
bool checkTelephone(string telephone){
    bool correct=true;
    
    if(telephone.length()<11 || telephone.length()>13 || telephone[0]!='+'){
        correct=false;
    }
    else{
        for(unsigned int i=1;i<telephone.length() && correct;i++){
            if(!isdigit(telephone[i])){
                correct=false;
            }
        }
    }
    
    return correct;
}

/*
Función que busca un paciente en la base de datos
data: base de datos
nif: NIF del paciente a buscar
return: posición del paciente en la base de datos o -1 si no existe
*/
int searchPatient(const Database &data,string nif){
    int pos=-1;

    for(unsigned int i=0;i<data.patients.size() && pos==-1;i++){
        if(data.patients[i].nif==nif){
            pos=i;
        }
    }

    return pos;
}

/*
Función que añade un paciente a la base de datos
data: base de datos
return: void
*/
void addPatient(Database &data){
    Patient p;
    bool nifCorrect,nameCorrect,telephoneCorrect;
    nifCorrect=nameCorrect=telephoneCorrect=false;

    do{
        cout << "Enter NIF: ";
        getline(cin, p.nif);
        
        if(p.nif!=""){
            if(checkNif(p.nif)){
                if(searchPatient(data,p.nif)==-1){
                    nifCorrect=true;
                    do{
                        cout << "Enter name: ";
                        getline(cin, p.name);
                        if(checkName(p.name)){
                            nameCorrect=true;
                            do{
                                cout << "Enter telephone: ";
                                getline(cin, p.telephone);
                                if(checkTelephone(p.telephone)){
                                    telephoneCorrect=true;
                                    data.patients.push_back(p);
                                }
                                else{
                                    error(ERR_WRONG_TELEPHONE);
                                }
                            }while(!telephoneCorrect);
                        }
                        else{
                            error(ERR_WRONG_NAME);
                        }
                    }while(!nameCorrect);
                }
                else{
                    error(ERR_PATIENT_EXISTS);
                }
            }
            else{
                error(ERR_WRONG_NIF);
            }
        }
    }while(p.nif!="" && !nifCorrect);
}

/*
Función que da formato una fecha, añadiendo ceros a la izquierda si es necesario a día y mes
date: fecha a formatear
return: sgtring
*/
string formatDate(Date date){
    stringstream ss;

    if(date.day < 10){
        ss << '0';
    }
    ss << date.day << '/';
    if(date.month < 10){
        ss << '0';
    }
    ss << date.month << '/' << date.year;
    
    return ss.str();
}

/* 
Función que muestra los datos de un paciente
data: base de datos
return: void
*/
void viewPatient(const Database &data){
    string nif;
    int index=-1;
    bool header=false;

    do{
        cout << "Enter NIF: ";
        getline(cin, nif);

        if(nif!=""){
            if(checkNif(nif)){
                index=searchPatient(data,nif);
                if(index!=-1){
                    cout << "NIF: " << data.patients[index].nif << endl;
                    cout << "Name: " << data.patients[index].name << endl;
                    cout << "Telephone: " << data.patients[index].telephone << endl;
                    for(unsigned int i=0;i<data.analysis.size();i++){
                        if(data.analysis[i].nif==nif){
                            if(!header){
                                cout << "Id\tDate\tHeight\tWeight" << endl;
                                header=true;
                            }
                            cout << data.analysis[i].id << '\t';
                            cout << formatDate(data.analysis[i].dateAnalysis);
                            cout << '\t' << data.analysis[i].height << '\t' << data.analysis[i].weight << endl;
                        }
                    }
                }
                else{
                    error(ERR_PATIENT_NOT_EXISTS);
                }
            }
            else{
                error(ERR_WRONG_NIF);
            }
        }
    }while(index==-1 && nif!="");
}

/*
Función que elimina un paciente de la base de datos
data: base de datos
return: void
*/
void deletePatient(Database &data){
    string nif;
    int index=-1;

    do{
        cout << "Enter NIF: ";
        getline(cin, nif);

        if(nif!=""){
            if(checkNif(nif)){
                index=searchPatient(data,nif);
                if(index!=-1){
                    data.patients.erase(data.patients.begin()+index);
                    for(unsigned int i=0;i<data.analysis.size();i++){
                        if(data.analysis[i].nif==nif){
                            data.analysis.erase(data.analysis.begin()+i);
                            i--; // Para que no se salte la siguiente posición
                        }
                    }
                }
                else{
                    error(ERR_PATIENT_NOT_EXISTS);
                }
            }
            else{
                error(ERR_WRONG_NIF);
            }
        }
    }while(index==-1 && nif!="");
}

/*
Función que guarda los pacientes de la base de datos en fichero
data: base de datos
return: void
*/
void savePatients(const Database &data){
    ofstream file(KFILEPATIENTS,ios::out|ios::binary);

    if(file.is_open()){
        PatientBin patient;
        for(unsigned int i=0;i<data.patients.size();i++){
            // Copiar los datos de la estructura a la estructura binaria
            strcpy(patient.nif,data.patients[i].nif.c_str()); // Copiamos el NIF
            strncpy(patient.name,data.patients[i].name.c_str(),KMAXNAME-1); // Recortar el nombre por si es mayor de KMAXNAME-1
            patient.name[KMAXNAME-1]='\0'; // strncpy no añade el carácter nulo al final de la cadena
            strcpy(patient.telephone,data.patients[i].telephone.c_str()); // Copiamos el teléfono
            
            file.write((const char *)&patient,sizeof(PatientBin));
        }
        file.close();
    }
}

/*
Función que comprueba si una fecha es correcta
date: fecha a comprobar
return: true si es correcta, false en caso contrario
*/
bool checkDate(Date date){
    bool correct=false;

    if(date.day>=1 && date.day<=31 && date.month>=1 && date.month<=12 && date.year>=2025 && date.year<=2050){
        correct=true;
    }
    
    return correct;
}

/*
Función que añade una analítica a un paciente
data: base de datos
return: void
*/
void addAnalysis(Database &data){
    Analysis newAnalysis;
    string nif;
    int index=-1;
    bool dateCorrect, heightCorrect, weightCorrect;
    dateCorrect=heightCorrect=weightCorrect=false;

    do{
        // Pedimos el NIF
        cout << "Enter NIF: ";
        getline(cin, nif);
        if(nif!=""){
            index=searchPatient(data,nif);
            if(index!=-1){
                strcpy(newAnalysis.nif,nif.c_str());
                // Pedimos la fecha
                do{
                    dateCorrect=true;
                    cout << "Enter date (day/month/year): ";
                    cin >> newAnalysis.dateAnalysis.day;
                    cin.ignore(); // Descartamos el separador '/'
                    cin >> newAnalysis.dateAnalysis.month;
                    cin.ignore(); // Descartamos el separador '/'
                    cin >> newAnalysis.dateAnalysis.year;
                    cin.ignore(); // Descartamos el salto de línea
                    if(!checkDate(newAnalysis.dateAnalysis)){
                        error(ERR_WRONG_DATE);
                        dateCorrect=false;
                    }
                }while(!dateCorrect);
                // Pedimos el peso
                do{
                    weightCorrect=true;
                    cout << "Enter weight: ";
                    cin >> newAnalysis.weight;
                    cin.ignore();
                    if(newAnalysis.weight<0){
                        error(ERR_WRONG_NUMBER);
                        weightCorrect=false;
                    }
                }while(!weightCorrect);
                // Pedimos la altura
                do{
                  heightCorrect=true;
                  cout << "Enter height: ";
                  cin >> newAnalysis.height;
                  cin.ignore();
                  if(newAnalysis.height<0){
                      error(ERR_WRONG_NUMBER);
                      heightCorrect=false;
                  }
                }while(!heightCorrect);
                newAnalysis.id=data.nextId++; // Asignamos el siguiente identificador
                data.analysis.push_back(newAnalysis);
            }
            else{
                error(ERR_PATIENT_NOT_EXISTS);
            }
        }
    }while(index==-1 && nif!="");
}

/*
Función que exporta las analíticas de la base de datos a fichero binario
data: base de datos
return: void
*/
void exportAnalysis(const Database &data){
    string nif;

    ofstream file(KFILEANALYSIS,ios::out|ios::binary);

    if(file.is_open()){
        for(unsigned int i=0;i<data.analysis.size();i++){
            file.write((const char *)&data.analysis[i],sizeof(Analysis));
        }

        file.close();
    }
    else{
        error(ERR_FILE);
    }
}

/*
Función que importa las analíticas de un fichero binario a la base de datos
data: base de datos
return: void
*/
void importAnalysis(Database &data){
    ifstream fileAnalysis(KFILEANALYSIS,ios::in|ios::binary);
    ofstream fileWrong(KFILEWRONGPATIENTS,ios::out|ios::app); // Añadir al final del fichero

    if(fileAnalysis.is_open() && fileWrong.is_open()){
        Analysis newAnalysis;

        while(fileAnalysis.read((char *)&newAnalysis,sizeof(Analysis))){
            if(searchPatient(data,newAnalysis.nif)!=-1){
                newAnalysis.id=data.nextId++;
                data.analysis.push_back(newAnalysis);
            }
            else{
                fileWrong << newAnalysis.nif << endl;
            }
        }

        fileAnalysis.close();
        fileWrong.close();
    }
    else{
        error(ERR_FILE);
    }
}

/*
Función que calcula el IMC de una analítica
a: analítica a calcular
return: float
*/
string calculateIMC(const Analysis &a){
    float imc;
    string type;

    imc=a.weight/(a.height/100*a.height/100);

    if(imc<18.5){
        type="Underweight";
    }
    else if(imc>=18.5 && imc<25){
        type="Healthy";
    }
    else if(imc>=25 && imc<30){
        type="Overweight";
    }
    else{
        type="Obesity";
    }

    return type;
}

/*
Función que muestra las estadísticas de las analíticas
data: base de datos
return: void
*/
void statistics(const Database &data,bool save=true){
    ofstream fileStatistics(KFILESTATISTICS,ios::out);

    if (fileStatistics.is_open()){
        for(unsigned int i=0;i<data.analysis.size();i++){
            // Mostar por pantalla
            cout << data.analysis[i].nif << ";" << formatDate(data.analysis[i].dateAnalysis) << ";" << data.analysis[i].weight << ";" << data.analysis[i].height << ";" << calculateIMC(data.analysis[i]) << endl;
            // Guardar en fichero
            if(save){
                fileStatistics << data.analysis[i].nif << ";" << formatDate(data.analysis[i].dateAnalysis) << ";" << data.analysis[i].weight << ";" << data.analysis[i].height << ";" << calculateIMC(data.analysis[i]) << endl;
            }
        }
        fileStatistics.close();
    }
    else{
        error(ERR_FILE);
    }
}

/*
Función que comprueba los argumentos pasados por línea de comandos
argc: número de argumentos
argv: argumentos pasados al programa
fileName: nombre del fichero de analíticas
showStatistics: mostrar estadísticas
return: true si los argumentos son correctos, false en caso contrario
*/
bool checkArguments(int argc, char *argv[], string &fileName, bool &showStatistics){
    bool correct=true;
    bool existF=false;
    bool existS=false;

    for (int i=1;i<argc && correct;i++){
        if(!existF && strcmp(argv[i],"-f")==0){
            if(i+1<argc){
                existF=true;
                fileName=argv[i+1];
                i++;
            }
            else{
                correct=false;
            }
        }
        else if(!existS && strcmp(argv[i],"-s")==0){
            existS=true;
            showStatistics=true;
        }
        else{
            correct=false;
        }
    }
    
    if(existS && !existF){
        correct=false;
    }

    return correct;
}

/*
Función que lee las analíticas de un fichero de texto pasado como argumento y las añade a la base de datos
data: base de datos
fileName: nombre del fichero
return: void
*/
void readAnalysis(Database &data,string fileName){
    ifstream fileAnalysis(fileName,ios::in);
    ofstream fileWrong(KFILEWRONGPATIENTS,ios::out|ios::app); // Añadir al final del fichero
    string line;
    Analysis newAnalysis;

    if(fileAnalysis.is_open() && fileWrong.is_open()){
        while(getline(fileAnalysis,line)){
            stringstream ss(line); // Convertimos la línea en un flujo de texto
            string nif;
            Date date;
            float weight,height;

            getline(ss,nif,';'); // Leemos el NIF (hasta el primer ';')
            if(searchPatient(data,nif)!=-1){
                newAnalysis.id=data.nextId++;
                strcpy(newAnalysis.nif,nif.c_str());
                ss >> date.day;
                ss.ignore(); // Descartamos el separador '/'
                ss >> date.month;
                ss.ignore(); // Descartamos el separador '/'
                ss >> date.year;
                ss.ignore(); // Descartamos el ';'
                newAnalysis.dateAnalysis=date;
                ss >> weight;
                newAnalysis.weight=weight;
                ss.ignore(); // Descartamos el ';'
                ss >> height;
                newAnalysis.height=height;
                data.analysis.push_back(newAnalysis);
            }
            else{
                fileWrong << nif << endl;
            }
        }

        fileAnalysis.close();
        fileWrong.close();
    }
    else{
        error(ERR_FILE);
    }
}

/*
Función que añade los pacientes en fichero a la base de datos
data: base de datos
return: void
*/
void loadPatients(Database &data){
    ifstream file(KFILEPATIENTS,ios::in|ios::binary);

    if(file.is_open()){
        PatientBin patient;

        while(file.read((char *)&patient,sizeof(PatientBin))){
            Patient p;
            p.nif=patient.nif;
            p.name=patient.name;
            p.telephone=patient.telephone;
            data.patients.push_back(p);
        }

        file.close();
    }
}

/*
Función que muestra el menú de opciones
return: void
*/
void showMenu() {
    cout << "1- Add patient" << endl
         << "2- View Patient" << endl
         << "3- Delete patient" << endl
         << "4- Save patients" << endl
         << "5- Add analysis" << endl
         << "6- Export analysis" << endl
         << "7- Import analysis" << endl
         << "8- Statistics" << endl
         << "q- Quit" << endl
         << "Option: ";
}

/*
Función principal
return: 0
*/
int main(int argc, char *argv[]){
    Database data;
    data.nextId=1;
    char option;
    string fileName;
    bool showStatistics=false;

    if(checkArguments(argc,argv,fileName,showStatistics)){
         // Cargar los pacientes en fichero a la base de datos
        loadPatients(data);

        // Cargar las analíticas de un fichero de estadísticas
        if(fileName!=""){
            readAnalysis(data,fileName);
        }

        // Mostrar las estadísticas
        if(showStatistics){
            statistics(data,false); // Mostrar pero no guardar en fichero
        }
        else{
            do{
                showMenu();
                cin >> option;
                cin.ignore();
                
                switch(option){
                    case '1': addPatient(data);
                        break;
                    case '2': viewPatient(data);
                        break;
                    case '3': deletePatient(data);
                        break;
                    case '4': savePatients(data);
                        break;
                    case '5': addAnalysis(data);
                        break;
                    case '6': exportAnalysis(data);
                        break;
                    case '7': importAnalysis(data);
                        break;
                    case '8': statistics(data);
                        break;
                    case 'q': // Salir del programa 
                        break;
                    default: error(ERR_OPTION);
                }
            }while(option!='q');
        }
    }
    else{
        error(ERR_ARGS);
    }

    return 0;
  }